from .main import SOARException
from .validation_exc import (
    DataError,
    NoInputError,
    NoSecretsError,
    BadInputError,
    BadSecretsError
)

from .protocol_exc import (
    ProtocolError,
    ConnectionFailed,
    CredentialsError,
    PermissionsError,
    ExecutionError
)


from ussl.exceptions import SOARException


class DataError(SOARException):
    """ Базовый класс для ошибок валидации. """
    _return_code: int = 10
    _TEMPLATE = 'Data error: {args}.'


class NoInputError(DataError):
    """ Класс для ошибок отсутствия обязательных входных данных. """
    _return_code: int = 11
    _TEMPLATE = 'Missing input data: {args}.'


class NoSecretsError(DataError):
    """ Класс для ошибок отсутствия обязательных секретов. """
    _return_code: int = 12
    _TEMPLATE = 'Missing secrets: {args}.'


class BadInputError(DataError):
    """ Класс для ошибок неверных входных данных. """
    _return_code: int = 13
    _TEMPLATE = 'Input data validation error: {args}.'


class BadSecretsError(DataError):
    """ Класс для ошибок неверных секретов. """
    _return_code: int = 14
    _TEMPLATE = 'Secrets validation error: {args}.'



from typing import List, Union


class SOARException(Exception):
    """
    Стандартный класс для исключений скриптов.

    Приватные поля:
        ``_return_code``(int): код возврата
        ``_TEMPLATE``(str): шаблон сообщения

    Публичные поля и свойства:
        ``message``(Union[List[str], str, dict]): информация об исключении
        ``return_code``(int): код возврата
    """
    _return_code: int = 1
    _TEMPLATE: str = 'Error: {args}.'

    message: Union[List[str], str, dict] = None

    @property
    def return_code(self):
        return self._return_code

    def __init__(self, message: Union[List[str], str, dict]):
        self.message = message

    def __str__(self):
        if isinstance(self.message, List):
            return self._TEMPLATE.format(args=', '.join(self.message))
        elif isinstance(self.message, str):
            return self._TEMPLATE.format(args=self.message)
        else:
            return self._TEMPLATE.format(args=str(self.message))


#! /usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='ussl',
    author='ussc soc dev team',
    author_email='iushangaraev@ussc.ru, pbikkuzhina@ussc.ru',
    description='Пакет разработчиков USSC-SOC для упрощения взаимодействия с АРМ, серверами и сетевыми устройствами',
    long_description=long_description,
    long_description_content_type='text/markdown',
    version='1.0.14',
    packages=[
        'ussl',
        'ussl.model',
        'ussl.postprocessing',
        'ussl.protocol',
        'ussl.transport',
        'ussl.exceptions',
        'ussl.utils',
    ],
    install_requires=[
        'pywinrm==0.4.1',
        'paramiko==2.7.2',
        'marshmallow==3.20.2',
        'python-ldap==3.4.0', # адаптер https://www.lfd.uci.edu/~gohlke/pythonlibs/#python-ldap есть только для этой версии библиотеки
    ],
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.8'
    ]
)

# python setup.py sdist
# twine upload dist/*



import sys
import json
import warnings
import pathlib
from abc import abstractmethod, ABC
from json import JSONDecodeError
from typing import Union, Tuple, Type

from marshmallow import Schema, exceptions, EXCLUDE
from ussl.exceptions import (
    NoSecretsError,
    BadInputError,
    BadSecretsError,
    NoInputError,
    DataError,
    SOARException
)

warnings.filterwarnings("ignore")


class BaseFunction(ABC):
    """
    Является базовым классом для всех скриптов, участвующих в обогащении и реагировании.
    При использовании класса необходимо реализовать метод ``function``.
    Автоматически принимаемые значения:
        ``input_json``: Первым аргументом принимает информацию, переданную на вход плейбука;
        ``secrets``: Вторым аргументом приниает секреты.
        ``ensure_ascii``: Указывает, должны ли символы не из набора ASCII быть экранированы. По умолчанию False.
    """
    inputs_model: Type[Schema] = None
    secrets_model: Type[Schema] = None
    ensure_ascii: bool = False
    _input_json: dict = None
    _secrets: dict = None

    def __init__(self, ensure_ascii: bool = False) -> None:
        """
        Инициализирует экземпляр класса.
        Вызывает чтение и валидацию данных. А также, выполнение функции скрипта.
        Args:
            ``ensure_ascii (bool)``: Указывает, должны ли символы не из набора ASCII быть экранированы. По умолчанию False
        Returns:
            ``None``
        """
        self.ensure_ascii = ensure_ascii

        try:
            self._set_valid_input_data()
        except DataError as e:
            self._output_json(str(e), e.return_code)

        try:
            result, message = self.function()
        except SOARException as e:
            self._output_json(str(e), e.return_code)
        except (TypeError, NotImplementedError):
            self._output_json("Incorrect script", 1)
        else:
            self._output_json(message, **result)

    def _set_valid_input_data(self):
        # Читаем входные данные
        try:
            inputs = pathlib.Path(sys.argv[1]).read_text(encoding='utf-8')
            self._input_json = json.loads(inputs)
        except FileNotFoundError:
            raise NoInputError("Input file not found")
        except JSONDecodeError:
            raise BadInputError("Input data is not valid JSON")
        # Читаем секреты
        try:
            secrets = pathlib.Path(sys.argv[2]).read_text(encoding='utf-8')
            self._secrets = json.loads(secrets)['secrets']
        except FileNotFoundError:
            raise NoSecretsError("Secrets file not found")
        except JSONDecodeError:
            raise BadSecretsError("Secrets data is not valid JSON")
        except KeyError:
            raise NoSecretsError("Secrets not found")

        # Валидируем входные данные
        if self.inputs_model is not None:
            try:
                self.input_json = self.inputs_model(unknown=EXCLUDE).load(self._input_json)
            except exceptions.ValidationError as e:
                raise BadInputError(e.args[0])
        else:
            self.input_json: dict = self._input_json.copy()
        # Валидируем секреты
        if self.secrets_model is not None:
            try:
                self.secrets = self.secrets_model(unknown=EXCLUDE).load(self._secrets)
            except exceptions.ValidationError as e:
                raise BadSecretsError(e.args[0])
        else:
            self.secrets: dict = self._secrets.copy()

    @abstractmethod
    def function(self) -> Tuple[dict, str]:
        """
        В этом методе необходимо реализовать функцию по обогащению
        или реагированию.

        Методу доступны переменные input_json и secrets.

        Для получения данных используйте переменные input_json и secrets класса BaseFunction.
        Для вывода ошибок необходимо использовать исключения из модуля exceptions.
        Returns:
            (dict, str): Результат обогащения или реагирования и сообщение о результате.
        """
        raise NotImplementedError('Метод function не реализован')

    def _output_json(self,
                     message: Union[str, dict],
                     return_code: int = 0,
                     **kwargs) -> None:
        """
        Выводит результат работы скрипта в формате JSON.

        Args:
            ``message (str)``: Сообщение о результате выполнения скрипта, которое будет выведено.
            ``return_code (int)``: Код возврата, указывающий на успешное выполнение (0) или ошибку (ненулевое значение).
            ``**kwargs``: Дополнительные именованные аргументы. Например, результат сбора данных.

        Returns:
            ``None``
        """
        # Обновляем входной JSON с результатом или сообщением об ошибке
        self._input_json['error' if return_code else 'result'] = message

        # Обновляем входной JSON новыми аргументами
        self._input_json.update(kwargs)

        # Выводим входной JSON в форматированном виде
        print(json.dumps(self._input_json, ensure_ascii=self.ensure_ascii))

        # Завершаем выполнение скрипта с кодом 0 в случае успешного выполнения или ненулевым в случае ошибки
        exit(return_code)



import re
import ssl
from typing import Any, Union

import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import ReadTimeout
from urllib3 import ProxyManager
from urllib3.util.ssl_ import create_urllib3_context
from winrm.exceptions import (
    AuthenticationError,
    WinRMError,
    WinRMOperationTimeoutError,
    WinRMTransportError,
)
from winrm.protocol import Protocol as WmProtocol

from .base import BaseProtocol
from ..model import Response, Protocol, Query
from ..exceptions import ConnectionFailed, ExecutionError, CredentialsError

WINRM_PORT = 5985
WINRM_SSL_PORT = 5986
VALID_TRANSPORT = {'plaintext', 'ntlm'}


class TLSv1Adapter(HTTPAdapter):
    """
    Включаем поддержку TLSv1.0 для старых версий Windows.
    """

    def init_poolmanager(self, *args: Any, **kwargs: Any) -> None:
        context = create_urllib3_context()
        context.minimum_version = ssl.TLSVersion.TLSv1
        context.options &= (
            ~ssl.OP_NO_TLSv1_3  # pylint: disable=no-member
            & ~ssl.OP_NO_TLSv1_2
            & ~ssl.OP_NO_TLSv1_1
            & ~ssl.OP_NO_TLSv1
        )
        kwargs['ssl_context'] = context
        return super().init_poolmanager(*args, **kwargs)

    def proxy_manager_for(self, *args: Any, **kwargs: Any) -> ProxyManager:
        context = create_urllib3_context()
        context.minimum_version = ssl.TLSVersion.TLSv1
        context.options &= (
            ~ssl.OP_NO_TLSv1_3  # pylint: disable=no-member
            & ~ssl.OP_NO_TLSv1_2
            & ~ssl.OP_NO_TLSv1_1
            & ~ssl.OP_NO_TLSv1
        )
        kwargs['ssl_context'] = context
        return super().proxy_manager_for(*args, **kwargs)


class WinRMProtocol(BaseProtocol):
    name = 'winrm'

    def __init__(self) -> None:
        self._protocol: WmProtocol
        self._shell_id = None
        # self._encoding = 437

    def connect(
        self,
        protocol: Protocol,
    ) -> None:
        host = protocol.host
        username = protocol.username
        domain = protocol.domain
        password = protocol.password
        scheme = protocol.default('scheme', 'https')
        if scheme == 'https':
            port = protocol.default('port', WINRM_SSL_PORT)
        else:
            port = protocol.default('port', WINRM_PORT)
        path = protocol.default('path', 'wsman')
        # Отправка запроса в кодировке 437 даёт ответ на анлийском языке
        # однако при преобразовании байтов в строку и использованием той же
        # кодировки русские символы отображаются некорректно. Для того чтобы
        # обойти эту проблему, для декодирования используется кодировка 866
        encoding = protocol.default('encoding', 437)
        self._decoding = protocol.default('decoding', 866)
        window_width = protocol.default('window_width', 300)
        transport = protocol.default('transport', 'ntlm')

        if transport not in VALID_TRANSPORT:
            raise ConnectionFailed(f'{transport} protocol is not supported')

        # для BasicAuth явно преобразуем 'username' и 'password' из utf8 в cp1251, т.к. requests.auth.HTTPBasicAuth
        # по умолчанию преобразует utf8 строки в latin1, при этом bytes оставляет 'как есть'.
        # https://github.com/requests/requests/pull/3673
        _username: Union[str, bytes] = f'{domain}\\{username}'
        _password: Union[str, bytes] = password
        if transport == 'plaintext':
            _username = f'{domain}\\{username}'.encode('cp1251')
            _password = password.encode('cp1251')
        else:
            _username = f'{domain}\\{username}'
            _password = password

        endpoint = f'{scheme}://{host}:{port}/{path}'
        try:
            self._protocol = WmProtocol(
                endpoint=endpoint,
                username=_username,
                password=_password,
                transport=transport,
                server_cert_validation='ignore'
            )

            self._protocol.transport.build_session()
            self._protocol.transport.session.mount(endpoint, TLSv1Adapter())
            self._shell_id = self._protocol.open_shell(codepage=encoding)

            self.execute(Query(command=f'mode con:cols={window_width}', timeout=10))
        except requests.exceptions.ConnectionError as e:
            raise ConnectionFailed(str(e))

    def close(self) -> None:
        self._protocol.close_shell(self._shell_id)
        self._shell_id = None

    def execute(
            self,
            query: Query
            ) -> Response:
        self._protocol.read_timeout_sec = (
            self._protocol.transport.read_timeout_sec
        ) = query.timeout if query.timeout else 10

        try:
            command_id = self._protocol.run_command(self._shell_id, query.command)
            std_out, std_err, status_code = self._protocol.get_command_output(
                self._shell_id, command_id
            )
            if status_code != 0:
                std_err = std_err.decode(encoding=str(self._decoding)).strip()
                return Response(
                    result=std_err,
                    status=False,
                    text=f'Команда {query.command} не выполнена')

            std_out = std_out.decode(encoding=str(self._decoding)).strip()
            if query.expects is not None:
                for pattern in query.expects:
                    if re.search(pattern, std_out):
                        return Response(
                            result=std_out,
                            text=f'Команда {query.command} выполнена успешно',
                            status=True)
                return Response(
                    result=std_out,
                    text=f'Команда {query.command} вернула неожиданный ответ',
                    status=False)
            else:
                return Response(
                    result=std_out,
                    text=f'Команда {query.command} выполнена успешно',
                    status=True)

        except WinRMTransportError as exc:
            raise ConnectionFailed(exc.__str__())

        except AuthenticationError as exc:
            raise CredentialsError(exc.__str__())

        except (WinRMOperationTimeoutError, ReadTimeout):
            raise ConnectionFailed('Connection timeout')

        except WinRMError as exc:
            raise ExecutionError(
                {"command": query.command, "error": exc.__str__()}
            ) from exc

        finally:
            self._protocol.cleanup_command(self._shell_id, command_id)



from ussl.exceptions import SOARException


class ProtocolError(SOARException):
    """ Базовый класс, описывающий ошибки удаленного доступа. """
    _return_code: int = 20
    _TEMPLATE = 'Protocol error: {args}.'


class ConnectionFailed(ProtocolError):
    """ Класс, описывающий ошибку подключения к удаленному хосту. """
    _return_code: int = 21
    _TEMPLATE = 'Connection failed: {args}.'


class CredentialsError(ProtocolError):
    """ Класс, описывающий ошибку авторизации. """
    _return_code: int = 22
    _TEMPLATE = 'Invalid credentials: {args}.'


class PermissionsError(ProtocolError):
    """ Класс, описывающий ошибку доступа. """
    _return_code: int = 23
    _TEMPLATE = 'Missing permissions: {args}.'


class ExecutionError(ProtocolError):
    """ Класс, описывающий ошибку выполнения команд. """
    _return_code: int = 24
    _TEMPLATE = 'Command execution error: {args}.'
